/* Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/**
  @defgroup dtm_standalone main.c
  @{
  @ingroup ble_sdk_app_dtm_serial
  @brief Stand-alone DTM application for UART interface.

 */

#include <stdint.h>
#include <stdbool.h>
#include "nrf.h"
#include "ble_dtm.mod.h"
#include "bsp.h"
#include "nrf_gpio.h"
#include "nrf_delay.h"

uint32_t    current_time = 0;
uint32_t    msb_time          = 0;     // Time when MSB of the DTM command was read. Used to catch stray bytes from "misbehaving" testers.
bool        is_msb_read       = false; // True when MSB of the DTM command has been read and the application is waiting for LSB.
uint16_t    dtm_cmd_from_uart = 0;     // Packed command containing command_code:freqency:length:payload in 2:6:6:2 bits.
uint8_t     rx_byte;                   // Last byte read from UART.
dtm_event_t result;                    // Result of a DTM operation.

// @note: The BLE DTM 2-wire UART standard specifies 8 data bits, 1 stop bit, no flow control.
//        These parameters are not configurable in the BLE standard.

/**@details Maximum iterations needed in the main loop between stop bit 1st byte and start bit 2nd
 * byte. DTM standard allows 5000us delay between stop bit 1st byte and start bit 2nd byte.
 * As the time is only known when a byte is received, then the time between between stop bit 1st
 * byte and stop bit 2nd byte becomes:
 *      5000us + transmission time of 2nd byte.
 *
 * Byte transmission time is (Baud rate of 19200):
 *      10bits * 1/19200 = approx. 520 us/byte (8 data bits + start & stop bit).
 *
 * Loop time on polling UART register for received byte is defined in ble_dtm.c as:
 *   UART_POLL_CYCLE = 260 us
 *
 * The max time between two bytes thus becomes (loop time: 260us / iteration):
 *      (5000us + 520us) / 260us / iteration = 21.2 iterations.
 *
 * This is rounded down to 21.
 *
 * @note If UART bit rate is changed, this value should be recalculated as well.
 */
#define MAX_ITERATIONS_NEEDED_FOR_NEXT_BYTE ((5000 + 2 * UART_POLL_CYCLE) / UART_POLL_CYCLE)

#define TX_REMAP_PIN_NUMBER 12
#define RX_REMAP_PIN_NUMBER 14

/**@brief Function for UART initialization.
 */
static void uart_init(void)
{
    // Configure UART0 pins.
    nrf_gpio_cfg_output(TX_REMAP_PIN_NUMBER);
    nrf_gpio_cfg_input(RX_REMAP_PIN_NUMBER, NRF_GPIO_PIN_NOPULL);

    NRF_UART0->PSELTXD       = TX_REMAP_PIN_NUMBER;
    NRF_UART0->PSELRXD       = RX_REMAP_PIN_NUMBER;
    NRF_UART0->BAUDRATE      = DTM_BITRATE;

    // Clean out possible events from earlier operations
    NRF_UART0->EVENTS_RXDRDY = 0;
    NRF_UART0->EVENTS_TXDRDY = 0;
    NRF_UART0->EVENTS_ERROR  = 0;

    // Activate UART.
    NRF_UART0->ENABLE        = UART_ENABLE_ENABLE_Enabled;
    NRF_UART0->INTENSET      = 0;
    NRF_UART0->TASKS_STARTTX = 1;
    NRF_UART0->TASKS_STARTRX = 1;
}

/**@brief Function for sending 2 bytes out through UART
 */
static void uart_send_16bit(uint16_t twobyte)
{
    // Report command status on the UART.
    // Transmit MSB of the result.
    NRF_UART0->TXD = (twobyte >> 8) & 0xFF;
    // Wait until MSB is sent.
    while (NRF_UART0->EVENTS_TXDRDY != 1)
    {
        // Do nothing.
    }
    NRF_UART0->EVENTS_TXDRDY = 0;
    
    // Transmit LSB of the result.
    NRF_UART0->TXD = twobyte & 0xFF;
    // Wait until LSB is sent.
    while (NRF_UART0->EVENTS_TXDRDY != 1)
    {
        // Do nothing.
    }
    NRF_UART0->EVENTS_TXDRDY = 0;
}

/**@brief Function for receiving 2 bytes from UART with timeout
 */
static uint32_t uart_receive_16bit(uint16_t *twobyte)
{
    uint8_t rx_byte;
    uint16_t rx_short = 0;
    
    for (;;)
    {
        if (NRF_UART0->EVENTS_RXDRDY == 0)
        {
            // Nothing read from the UART.
            continue;
        }
        NRF_UART0->EVENTS_RXDRDY = 0;
        rx_byte                  = (uint8_t)NRF_UART0->RXD;

        if (!is_msb_read)
        {
            // This is first byte of two-byte command.
            is_msb_read       = true;
            rx_short = ((dtm_cmd_t)rx_byte) << 8;

            // Go back and wait for 2nd byte of command word.
            continue;
        }

        // 2-byte received from UART
        is_msb_read = false;   
        *twobyte = rx_short | rx_byte;
        break;
    }   
    
    return 0;
}
    
/**@brief Function for splitting UART command bit fields into separate command parameters for the DTM library.
*
 * @param[in]   command   The packed UART command.
 * @return      result status from dtmlib.
 */
static uint32_t dtm_cmd_put(uint16_t command)
{
    dtm_cmd_t      command_code = (command >> 14) & 0x03;
    dtm_freq_t     freq         = (command >> 8) & 0x3F;
    uint32_t       length       = (command >> 2) & 0x3F;
    dtm_pkt_type_t payload      = command & 0x03;

    // Check for Vendor Specific payload.
    if (payload == 0x03)
    {
        /* Note that in a HCI adaption layer, as well as in the DTM PDU format,
           the value 0x03 is a distinct bit pattern (PRBS15). Even though BLE does not
           support PRBS15, this implementation re-maps 0x03 to DTM_PKT_VENDORSPECIFIC,
           to avoid the risk of confusion, should the code be extended to greater coverage.
        */
        payload = DTM_PKT_VENDORSPECIFIC;
    }
    return dtm_cmd(command_code, freq, length, payload);
}

#define TX_PASS_PIN_NUMBER 21
#define TX_FAIL_PIN_NUMBER 22
#define RX_PASS_PIN_NUMBER 23
#define RX_FAIL_PIN_NUMBER 24

/**@brief Function for initializing stuffs for the production test mode
 *
 * @return      result status from dtmlib.
 */
static uint32_t ptm_tester_init(void)    
{
    nrf_gpio_cfg_output(TX_PASS_PIN_NUMBER);
    nrf_gpio_cfg_output(TX_FAIL_PIN_NUMBER);
    nrf_gpio_cfg_output(RX_PASS_PIN_NUMBER);
    nrf_gpio_cfg_output(RX_FAIL_PIN_NUMBER);

    nrf_gpio_pin_clear(TX_PASS_PIN_NUMBER);
    nrf_gpio_pin_clear(TX_FAIL_PIN_NUMBER);
    nrf_gpio_pin_clear(RX_PASS_PIN_NUMBER);
    nrf_gpio_pin_clear(RX_FAIL_PIN_NUMBER);
}

//// DTM Command Packet Format ///  
// CC FFFFFF LLLLLL PP
// CC - Command (2-bit)
// FFFFFF - Frequency/Control (6-bit)
// LLLLLL - Length/Parameters (6-bit)
// PP - Packet Type for TX test (2-bit)
#define DTM_CMD(CMD,CTRL,LEN,PTYPE) ((CMD<<14)|(CTRL<<8)|(LEN<<2)|PTYPE)

#define DTM_CMD_RESET   0x0
#define DTM_CMD_RX      0x1
#define DTM_CMD_TX      0x2
#define DTM_CMD_TESTEND 0x3

#define DTM_CMD_PTYPE_PRBS     0
#define DTM_CMD_PTYPE_11110000 1
#define DTM_CMD_PTYPE_10101010 2
#define DTM_CMD_PTYPE_VS       3

static uint16_t test_packet_num [] = { 1, 100, 500, 1000};

/**@brief Function for going through a DUT rx test
 *
 * @param[in]   command   The packed UART command.
 * @return      result status from dtmlib.
 */
#define RX_TEST_PACKET_NUM_OPT 2
#define RX_TEST_PACKET_LEN     16
#define RX_TEST_CHANNEL        19
static uint32_t perform_dut_rx_test(void)    
{
    uint16_t command;
    uint16_t response;
    uint32_t marked_time;
    
    //// DTM Command Packet Format ///  
    // CC FFFFFF LLLLLL PP
    // CC - Command (2-bit)
    // FFFFFF - Frequency/Control (6-bit)
    // LLLLLL - Length/Parameters (6-bit)
    // PP - Packet Type for TX test (2-bit)
  
    //// RESET(DTM) ////
    command = DTM_CMD_RESET;
    uart_send_16bit(command);
    
    if( uart_receive_16bit(&response) == 0)
    {
        if(response == 0x0000)
        {
            result = 1;
        }
    }

    nrf_delay_ms(1);

    //// Put DUT in RX mode (DTM) ///
    //// Channel = 19
    command = DTM_CMD(DTM_CMD_RX,RX_TEST_CHANNEL,RX_TEST_PACKET_LEN,DTM_CMD_PTYPE_PRBS); // 0x5340
    uart_send_16bit(command);
    
    if( uart_receive_16bit(&response) == 0)
    {
        if(response == 0x0000)
        {
            result = 1;
        }
    }  

    nrf_delay_ms(1);
    
    //// Configure the number of packet to be sent by the TESTER (local) ////
    // CMD = TX
    // Packet number = (0 - 1, 1 - 100, 2 - 500, 3 - 1000, 4 - 10000) 
    // VS command = 4 (Set packet number)
    // Packet type = VS
    command = DTM_CMD(DTM_CMD_TX,RX_TEST_PACKET_NUM_OPT,4,DTM_CMD_PTYPE_VS); // 0x8213
    dtm_cmd_put(command);

    if (dtm_event_get(&result))
    {
    }

    //// Put TESTER into TX mode (local) ////
    // CMD = TX
    // Channel = 19
    // Length = 16 byte
    // Packet Type = PRBS
    command = DTM_CMD(DTM_CMD_TX,RX_TEST_CHANNEL,RX_TEST_PACKET_LEN,DTM_CMD_PTYPE_PRBS); // 0x9340
    dtm_cmd_put(command);

    if (dtm_event_get(&result))
    {
    }

    nrf_delay_ms(1);
    
    // It is roughly 1 packet sent per time mark
    marked_time = current_time;
    while( (current_time - marked_time) < test_packet_num[RX_TEST_PACKET_NUM_OPT])
    {
        current_time = dtm_wait();
    }

    nrf_delay_ms(1);
    
    //// TEST END(DTM) ////
    command = DTM_CMD(DTM_CMD_TESTEND,0,0,0);
    uart_send_16bit(command);   

    if( uart_receive_16bit(&response) == 0)
    {
        if( (response&0x3FFF) > test_packet_num[RX_TEST_PACKET_NUM_OPT])
        {
            if( ((response&0x3FFF))-test_packet_num[RX_TEST_PACKET_NUM_OPT] < 50 )
            {
                result = response;
            }
        }
        else
        {
            if( (test_packet_num[RX_TEST_PACKET_NUM_OPT]-(response&0x3FFF)) < 50 )
            {
                result = response;
            }
        }        
    }   

    if(result)
    {
        nrf_gpio_pin_clear(RX_PASS_PIN_NUMBER);
        nrf_gpio_pin_set(RX_FAIL_PIN_NUMBER);
    }
    else
    {
        nrf_gpio_pin_set(RX_PASS_PIN_NUMBER);
        nrf_gpio_pin_clear(RX_FAIL_PIN_NUMBER);
    }
}

/**@brief Function for going through a DUT tx test
*
 * @param[in]   command   The packed UART command.
 * @return      result status from dtmlib.
 */
#define TX_TEST_PACKET_NUM_OPT 2
#define TX_TEST_PACKET_LEN     16
#define TX_TEST_CHANNEL        19
static uint32_t perform_dut_tx_test(void)    
{
    uint16_t command = 0;
    uint16_t response = 0;
    uint32_t marked_time;

    //// RESET(DTM) ////
    command = DTM_CMD_RESET;
    uart_send_16bit(command);
    
    if( uart_receive_16bit(&response) == 0)
    {
        if(response == 0x0000)
        {
            result = 1;
        }
    }

    nrf_delay_ms(1);

    //// Put TESTER in RX mode (local) ///
    //// Channel = 19
    // CMD = RX
    // Channel = 19
    // Length = 16 byte
    // Packet Type = PRBS
    command = DTM_CMD(DTM_CMD_RX,TX_TEST_CHANNEL,TX_TEST_PACKET_LEN,DTM_CMD_PTYPE_PRBS); // 0x5340
    dtm_cmd_put(command);

    if (dtm_event_get(&result))
    {
    }

    nrf_delay_ms(1);

    //// Configure DUT number of packet to transmit per test (DTM) ///
    //// Channel = 19
    // CMD = TX
    // Packet number = (0 - 1, 1 - 100, 2 - 1000, 3 - 2000) 
    // VS command = 4 (Set packet number)
    // Packet type = VS
    command = DTM_CMD(DTM_CMD_TX,TX_TEST_PACKET_NUM_OPT,4,DTM_CMD_PTYPE_VS); // 0x8213
    uart_send_16bit(command);
    
    if( uart_receive_16bit(&response) == 0)
    {
        if(response == 0x0000)
        {
            result = 1;
        }
    }  
    
    nrf_delay_ms(1);

    //// Put TX in RX mode (DTM) ///
    //// Channel = 19
    // CMD = TX
    // Channel = 19
    // Length = 16
    // Packet type = PRBS
    command = DTM_CMD(DTM_CMD_TX,TX_TEST_CHANNEL,TX_TEST_PACKET_LEN,DTM_CMD_PTYPE_PRBS); // 0x9340
    uart_send_16bit(command);
    
    if( uart_receive_16bit(&response) == 0)
    {
        if(response == 0x0000)
        {
            result = 1;
        }
    }  

    // It is roughly 1 packet receive per time mark
    marked_time = current_time;
    while( (current_time - marked_time) < 500)
    {
        current_time = dtm_wait();
    }

    nrf_delay_ms(1);

    //// Put TESTER into TEST END (local) ///
    //// Channel = 19
    command = DTM_CMD(DTM_CMD_TESTEND,0,0,0); // 0xC000
    dtm_cmd_put(command);

    if (dtm_event_get(&result))
    {
    }

    if( (result&0x3FFF) > test_packet_num[TX_TEST_PACKET_NUM_OPT])
    {
        if( ((result&0x3FFF)-test_packet_num[TX_TEST_PACKET_NUM_OPT]) < 50 )
        {
            result = 1;
        }
    }
    else
    {
        if( (test_packet_num[TX_TEST_PACKET_NUM_OPT]-(result&0x3FFF)) < 50 )
        {
            result = 1;
        }
    }
    
    if(result)
    {
        nrf_gpio_pin_clear(TX_PASS_PIN_NUMBER);
        nrf_gpio_pin_set(TX_FAIL_PIN_NUMBER);
    }
    else
    {
        nrf_gpio_pin_set(TX_PASS_PIN_NUMBER);
        nrf_gpio_pin_clear(TX_FAIL_PIN_NUMBER);
    }
}

/**@brief Function for application main entry.
 *
 * @details This function serves as an adaptation layer between a 2-wire UART interface and the
 *          dtmlib. After initialization, DTM commands submitted through the UART are forwarded to
 *          dtmlib and events (i.e. results from the command) is reported back through the UART.
 */
int main(void)
{
    uint32_t    dtm_error_code;

    uart_init();

    dtm_error_code = dtm_init();
    if (dtm_error_code != DTM_SUCCESS)
    {
        // If DTM cannot be correctly initialized, then we just return.
        return -1;
    }

    ptm_tester_init();  

    perform_dut_rx_test();
    
    perform_dut_tx_test();

    for (;;) {}
}

/// @}
